/* Program to transform words into likely phonetic forms, based on
   the common and consistent spellings listed in Calvert.

   Author: J. Zobel, jz@cs.rmit.edu.au, August 1995.
*/

#include <stdio.h>

#define BUFLEN 256

void	calvert(char *,char *);
void	initcalvert();

main(int argc, char *argv[])
{
    char	inbuf[BUFLEN], outbuf[BUFLEN];
    char	*gets();

    initcalvert();

    if( argc == 1 )
    {
	/* printf("wi> "); */
	while( gets(inbuf) != NULL )
	{
	    calvert(inbuf, outbuf);
	    printf("%s %s\n", inbuf, outbuf);
	    /* printf("wi> "); */
	}

	if( *inbuf == '\0' )
	    printf("\n");
    }
    else
    {
	calvert(argv[1], outbuf);
	printf("%s %s\n", argv[1], outbuf);
    }

    exit(0);
}


#define MAXSUBSTS	1000
#define SUBSTSFILE	"/usr/local/ipa/ipa/calvert.dfn"

typedef struct substrec
{
    char	*old, *new;
    int		l1, l2;
} SUBST;


static SUBST substs[MAXSUBSTS];
static int count;


/* Load a file of substitions into the array substs, adding + symbols.
   (These symbols allow disambiguation between characters and phonemes:
   characters have a preceding +.)  Any ^ or $ are put at the start of
   the match string, to allow easier processing later on */
   
void
initcalvert()
{
    FILE	*fp, *fopen();
    int		i, j, k, len;
    char	buf[BUFLEN], *fgets();
    char	*malloc();

    if( ( fp = fopen(SUBSTSFILE, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s\n", SUBSTSFILE);
	exit(1);
    }

    for( i=0 ; fgets(buf, BUFLEN, fp) != NULL ; i++ )
    {
	for( len=0 ; buf[len] != ' ' ; len++ )
	    ;
	substs[i].old = malloc( len*2 + 1 );
	if( buf[0] == '^' )
	{
	    substs[i].old[0] = '^';
	    j = k = 1;
	}
	else if( buf[len-1] == '$' )
	{
	    substs[i].old[0] = '$';
	    j = 0;
	    k = 1;
	    len--;
	}
	else
	    j = k = 0;

	for( ; j<len ; j++, k += 2 )
	{
	    substs[i].old[k] = '+';
	    substs[i].old[k+1] = buf[j];
	}

	if( buf[len] == '$' )
	     j++;
	j++;

	substs[i].old[k] = '\0';
	substs[i].l1 = k;

	substs[i].l2 = strlen( buf+j ) - 1;
	buf[j+substs[i].l2] = '\0';
	substs[i].new = malloc( substs[i].l2+1 );
	strcpy( substs[i].new, buf+j );

/* printf("(%s) (%s) %d %d\n", substs[i].old, substs[i].new, substs[i].l1, substs[i].l2); */
    }

    count = i;

    return;
}


/* Main transformation routine */
/* Assumes new always shorter than old */
/* Possible extension: treat hyphens as a special case, as delimiting
   separate sounds to be processed independently; eg tophat vs top-hat */
void
calvert(char *inbuf, char *outbuf)
{
    int		i, diff, off, k, len = strlen(inbuf);
    char	*strstr(), *strchr(), *p;

    for( i=0 ; i<len ; i++ )
    {
	outbuf[2*i] = '+';
	outbuf[2*i+1] = inbuf[i];
    }
    len = 2*i;
    outbuf[len] = '\0';

    for( k=0 ; k<count ; k++ )
    {
	if( substs[k].old[0] == '^' )
	{
	    if( strncmp(outbuf, substs[k].old+1, substs[k].l1-1) == 0 )
	    {
		diff = ( substs[k].l1-1 ) - substs[k].l2;
		len -= diff;
		for( i=0 ; i<substs[k].l2 ; i++ )
		    outbuf[i] = substs[k].new[i];
		for( ; outbuf[i] != '\0' ; i++ )
		    outbuf[i] = outbuf[i+diff];
	    }
	}
	else if( substs[k].old[0] == '$' )
	{
	    off = len - ( substs[k].l1 - 1 );
	    if( strcmp(outbuf+off, substs[k].old+1) == 0 )
	    {
		len -= ( ( substs[k].l1-1 ) - substs[k].l2 );
		strcpy( outbuf+off, substs[k].new );
	    }
	}
	else
	    while( ( p = strstr(outbuf, substs[k].old) ) != NULL )
	    {
		diff = ( substs[k].l1 ) - substs[k].l2;
		len -= diff;
		for( i=0 ; i<substs[k].l2 ; i++ )
		    p[i] = substs[k].new[i];
		for( ; p[i] != '\0' ; i++ )
		    p[i] = p[i+diff];
	    }
    }

    if( strchr(outbuf, '+') != NULL )
    {
	fprintf(stderr, "Warning: transformation %s -> %s incomplete\n",
	    inbuf, outbuf);
    }

    return;
}
