/* Edit distance routines.  Justin Zobel, jz@cs.rmit.edu.au */

#define true		1
#define false		0

#define STRLEN		60

int	edit3(char *, char *);
int	edit3taper(char *, char *);
int	edit4(char *, char *);

static char equal(char,char);
static int min3(int,int,int);
static int min4(int,int,int,int);

static int init=false;


/* Minimal edit distance: cost of insertion, deletion, or miss is 1 */
/* From Hall and Dowling */
int
edit3(char *in1, char *in2)
{
    int		F[STRLEN+1][STRLEN+1];
    int		i, j, l1, l2;

    l1 = strlen(in1);
    l2 = strlen(in2);

    if( !init )
    {
	for( i=0 ; i<=STRLEN ; i++ )
	    F[0][i] = F[i][0] = i;
	init = true;
    }

    for( i=1 ; i<=l1 ; i++ )
	for( j=1 ; j<=l2 ; j++ )
	    F[i][j] = min3(F[i-1][j] + 1,
			      F[i][j-1] + 1,
			      F[i-1][j-1] + equal(in1[i-1], in2[j-1]));
    
    return((int) F[l1][l2]);
}


/* Tapered edit distance: cost of insertion, deletion depends on posn
   in string */
int
edit3taper(char *in1, char *in2)
{
    int		F[STRLEN+1][STRLEN+1];
    int		i, j, l1, l2;
    int		sum, maxl, maxp;	/* maxp is the max penalty */

    l1 = strlen(in1);
    l2 = strlen(in2);

    maxl = (l1>l2) ? l1 : l2;

    maxp = 2 * ( l1 + l2 );	/* ie, twice minimum penalty */

    for( sum=i=0 ; i<=maxl ; i++ )
    {
	F[0][i] = F[i][0] = sum;
	sum += maxp - i;
    }

    for( i=1 ; i<=l1 ; i++ )
	for( j=1 ; j<=l2 ; j++ )
	    F[i][j] = min3(F[i-1][j] + maxp - i - j,
		      F[i][j-1] + maxp - i - j,
		      F[i-1][j-1] +
			(( equal(in1[i-1], in2[j-1])==0 ) ? 0 : maxp - i - j ));
    
    return((int) F[l1][l2]);
}


int
min3(int a, int b, int c)
{
    if( a > b )
	return( ( c > b ) ? b : c );
    else
	return( ( c > a ) ? a : c );
}


/* Minimal edit distance: cost of insertion, deletion, swap is 1 */
/* From Hall and Dowling */
int
edit4(char *in1, char *in2)
{
    int	F[STRLEN+1][STRLEN+1];
    int		i, j, l1, l2;

    l1 = strlen(in1);
    l2 = strlen(in2);

    if( !init )
    {
	for( i=0 ; i<=STRLEN ; i++ )
	    F[0][i] = F[i][0] = i;
	init = true;
    }

    for( i=1 ; i<=l1 ; i++ )
	for( j=1 ; j<=l2 ; j++ )
	{
	    if( i==1 || j==1 )
		F[i][j] = min4(F[i-1][j] + 1,
			      F[i][j-1] + 1,
			      F[i-1][j-1] + equal(in1[i-1], in2[j-1]) ,
			      1000000 );
	    else
		F[i][j] = min4(F[i-1][j] + 1,
			      F[i][j-1] + 1,
			      F[i-1][j-1] + equal(in1[i-1], in2[j-1]),
			      F[i-2][j-2] + equal(in1[i-2], in2[j-1])
					  + equal(in1[i-1], in2[j-2]) + 1);
	}
    
    return((int) F[l1][l2]);
}


int
min4(int a, int b, int c, int d)
{
    int		t;

    if( c > d )
	c = d;
    if( b > c )
	b = c;

    return( ( a > b ) ? b : a );
}


char
equal(char c1, char c2)
{
    return( ( c1 == c2 ) ? 0 : 1 );
}
