/* Use edit distance to join one file to another.  More efficient if
   the short file is given first.  Assumes each file is in the format

    key matchstring wgt

   Author: J. Zobel, jz@cs.rmit.edu.au
*/

#include <stdio.h>
#include <string.h>

#define true		1
#define false		0

#define STRLEN		256

#define USEWGT	false

#define EDIT3		0
#define EDIT3TAPER	1
#define IPA	 	2
#define IPATAPER 	3

int edittype = EDIT3;

#if USEWGT
void	doeditjoin(FILE*,char *,double);
#else
void	doeditjoin(FILE*,char *);
#endif


main(int argc, char *argv[])
{
    FILE	*f1, *f2, *fopen();
    char	buf[STRLEN], *sound, *p;
    char	*strchr(), *fgets();
#if USEWGT
    double	atof(), wgt;
#endif
    void	initipadist();

    if( argc == 4 )
    {
	if( strcmp(argv[1], "edit3") == 0 )
	    edittype = EDIT3;
	else if( strcmp(argv[1], "edit3taper") == 0 )
	    edittype = EDIT3TAPER;
	else if( strcmp(argv[1], "ipa") == 0 )
	    edittype = IPA;
	else if( strcmp(argv[1], "ipataper") == 0 )
	    edittype = IPATAPER;
	else
	{
	    fprintf(stderr, "Usage: editjoin [ edit3 | edit3taper | ipa | ipataper ] shortfile longfile\n");
	    exit(1);
	}

	argc--, argv++;
    }

    if( argc != 3 )
    {
	fprintf(stderr, "Usage: editjoin [ edit3 | edit3taper | ipa ] shortfile longfile\n");
	exit(1);
    }

    if( ( f1 = fopen(argv[1], "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading\n", argv[1]);
	exit(1);
    }
    if( ( f2 = fopen(argv[2], "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading\n", argv[2]);
	exit(1);
    }

    if( edittype == IPA || edittype == IPATAPER)
	initipadist();

	/* BUG assumes format of file is ok */
    while( fgets(buf, STRLEN, f1) != NULL )
    {
	sound = strchr(buf, (int) ' ');
	*sound = '\0';
	sound++;
#if USEWGT
	p = strchr(sound, (int) ' ');
	*p = '\0';
	p++;
	wgt = atof(p);
#else
	sound[strlen(sound)-1] = '\0';
#endif

#if USEWGT
	doeditjoin(f2, sound, wgt);
#else
	doeditjoin(f2, sound);
#endif
    }

    fclose(f1);
    fclose(f2);

    exit(0);
}


void
#if USEWGT
doeditjoin(FILE *f2, char *s1, double w1)
#else
doeditjoin(FILE *f2, char *s1)
#endif
{
    char	buf[STRLEN], *s2, *p;
    char	*strchr(), *fgets();
    int		len = strlen(s1);
    double	wgt;
#if USEWGT
    double	atof(), w2;
#endif
    int		edit3(char *, char *);
    int		ipadist(char *, char *);

	/* BUG assumes format of file is ok */
    while( fgets(buf, STRLEN, f2) != NULL )
    {
	s2 = strchr(buf, (int) ' ');
	*s2 = '\0';
	s2++;
#if USEWGT
	p = strchr(s2, (int) ' ');
	*p = '\0';
	p++;
	w2 = atof(p);
#else
	s2[strlen(s2)-1] = '\0';
#endif

	if( edittype == EDIT3 )
	{
#if USEWGT
	    wgt = w1 * w2 * ( 2 - edit3(s1, s2) ); /* allow one error */
#else
	    wgt = 6 - edit3(s1, s2); /* allow five!? errors */
#endif
	}
	else if( edittype == EDIT3TAPER )
	{
#if USEWGT
	    wgt = w1*w2 * ( 2*( strlen(s1)+strlen(s2) ) - edit3taper(s1, s2) );
#else
	    wgt = 50 - edit3taper(s1, s2);
#endif
	}
	else if( edittype==IPA )
	{
#if USEWGT
	    wgt = w1 * w2 * ( 10 - ipadist(s1, s2) ); /* allow one error */
#else
	    wgt = 100 - ipadist(s1, s2); /* allow about three errors */
#endif
	}
	else /* edittype == IPATAPER */
	{
#if USEWGT
	    wgt = w1 * w2 * ( 200 - ipadisttaper(s1, s2) );
#else
	    wgt = 1000 - ipadisttaper(s1, s2); /* allow about three errors */
#endif
	}

	if( wgt > 0 ) /* BUG highly arbitrary choice */
	    printf("%s %1.8f\n", buf, wgt);
    }
    return;
}
