/* Code for edit distance on phonemes.  Author J. Zobel, jz@cs.rmit.edu.au */

#include <stdio.h>

#define true		1
#define false		0

#define STRLEN		256
#define PHONEMEFILE	"/usr/local/ipa/ipa/phonemes.dfn"

#define HIGHDIST	7
#define MEDDIST		6
#define LOWDIST		4
#define MAXDIST		9

void	initipadist();
int	ipadist(char *,char *);
int	ipadisttaper(char *,char *);
static char phonemerep(char,char);
static char phonemedel(char,char);
static int min3(int,int,int);

typedef struct phonemerec
{
    char	high, low, med;
} PHONEME;

static PHONEME phonemes[256];

static int F[STRLEN+1][STRLEN+1];


void
initipadist()
{
    FILE	*fp, *fopen();
    int		i;
    char	buf[100], *fgets();

    if( ( fp = fopen(PHONEMEFILE, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading\n", PHONEMEFILE);
	exit(1);
    }

    for( i=0 ; i<256 ; i++ )
	phonemes[i].high = phonemes[i].med = phonemes[i].low = 0;
    
    while( fgets(buf, 100, fp) != NULL )
	    /* BUG doesn't check input format */
    {
	phonemes[buf[0]].high = buf[2];
	phonemes[buf[0]].med = buf[4];
	phonemes[buf[0]].low = buf[6];
    }

    return;
}


#define PHONEMESTART(c)	( ( phonemes[c].high == 'x' ) ? LOWDIST : MAXDIST )


/* Edit distance on phonemes, based on characteristics in PHONEMEFILE. */
int
ipadist(char *in1, char *in2)
{
    int		i, j, l1, l2;
    int		sum;
    int		up, left, diag;

    l1 = strlen(in1);
    l2 = strlen(in2);

    F[0][0] = 0;
    sum = PHONEMESTART(in1[0]);
    F[1][0] = sum;
    for( i=2 ; i<=l1 ; i++ )
    {
	sum += phonemedel(in1[i-2], in1[i-1]);
	F[i][0] = sum;
    }
    sum = PHONEMESTART(in2[0]);
    F[0][1] = sum;
    for( j=2 ; j<=l2 ; j++ )
    {
	sum += phonemedel(in2[j-2], in1[j-1]);
	F[0][j] = sum;
    }

    for( i=1 ; i<=l1 ; i++ )
    {
	for( j=1 ; j<=l2 ; j++ )
	{
	    up = ( i==1 ) ? PHONEMESTART(in1[0])
			    : phonemedel(in1[i-2], in1[i-1]);
	    left = ( j==1 ) ? PHONEMESTART(in2[0])
		            : phonemedel(in2[j-2], in2[j-1]);

	    diag = phonemerep(in1[i-1], in2[j-1]);

	    F[i][j] = min3(F[i-1][j] + up,
			      F[i][j-1] + left,
			      F[i-1][j-1] + diag);
	}
    }
    
    return((int) F[l1][l2]);
}


/* Edit distance on phonemes, based on characteristics in PHONEMEFILE. */
/* This routine does not behave well.  I suspect it is not working
   problem (or that the idea is stupid). */
int
ipadisttaper(char *in1, char *in2)
{
    int		i, j, l1, l2;
    int		maxp, sum;
    int		up, left, diag;

    l1 = strlen(in1);
    l2 = strlen(in2);

    maxp = ( l1+l2 ) * 2;

    F[0][0] = 0;
    sum = ( maxp-1 ) * PHONEMESTART(in1[0]);
    F[1][0] = sum;
    for( i=2 ; i<=l1 ; i++ )
    {
	sum += ( maxp - i ) * phonemedel(in1[i-2], in1[i-1]);
	F[i][0] = sum;
    }
    sum = ( maxp-1 ) * PHONEMESTART(in2[0]);
    F[0][1] = sum;
    for( j=2 ; j<=l2 ; j++ )
    {
	sum += ( maxp - j ) * phonemedel(in2[j-2], in1[j-1]);
	F[0][j] = sum;
    }

    for( i=1 ; i<=l1 ; i++ )
    {
	sum = maxp - i;
	for( j=1 ; j<=l2 ; j++ )
	{
	    sum--;
	    up = sum * ( ( i==1 ) ? PHONEMESTART(in1[0])
				    : phonemedel(in1[i-2], in1[i-1]) );
	    left = sum * ( ( j==1 ) ? PHONEMESTART(in2[0])
				    : phonemedel(in2[j-2], in2[j-1]) );

	    diag = sum * phonemerep(in1[i-1], in2[j-1]);

	    F[i][j] = min3(F[i-1][j] + up,
			      F[i][j-1] + left,
			      F[i-1][j-1] + diag);
	}
    }
    
    return((int) F[l1][l2]);
}


int
min3(int a, int b, int c)
{
    if( a > b )
	return( ( c > b ) ? b : c );
    else
	return( ( c > a ) ? a : c );
}


char
phonemerep(char c1, char c2)
{
    int		dist = 0;

    if( c1 == c2 )	/* logically unnecessary */
	return(0);

#define DEBUG 0
#if DEBUG
    if( phonemes[c1].high==0 )
	fprintf(stderr, "Warning, phoneme %c not known\n", c1);
    if( phonemes[c2].high==0 )
	fprintf(stderr, "Warning, phoneme %c not known\n", c2);
#endif /* DEBUG */

    if( phonemes[c1].high == 'x' && phonemes[c2].high == 'x' )
    {
#define ABS(a,b) ( (a>b)? a-b:b-a )
	dist = ABS( phonemes[c1].med, phonemes[c2].med )
		+ ABS( phonemes[c1].low, phonemes[c2].low );
    }
    else
    {
	if( phonemes[c1].high != phonemes[c2].high )
	    dist = HIGHDIST;
	if( phonemes[c1].med != phonemes[c2].med )
	    if( dist == 0 )
		dist = MEDDIST;
	    else
		dist++;
	if( phonemes[c1].low != phonemes[c2].low )
	    if( dist == 0 )
		dist = LOWDIST;
	    else
		dist++;
    }

    return(dist);
}


char
phonemedel(char c1, char c2)
{
    int		dist;

    if( phonemes[c1].high == 'x' && phonemes[c2].high == 'x' )
	dist = LOWDIST;
    else if( phonemes[c1].high == 'x' || phonemes[c2].high == 'x' )
	dist = MEDDIST;
    else
    {
	dist = phonemerep(c1, c2);
	if( dist < LOWDIST )
	    dist = LOWDIST;
    }

    return(dist);
}
