/* Compute recall-precision for `vrank -e' output.  Both
   vrank -e and relevance judgements must be of the form

   <qterm> <relterm> <relterm> ....

   The <qterm> is not assumed to be relevant.
*/

#include <stdio.h>

#define BUFLEN		1000000

typedef struct relrec
{
    char	*word;
    struct relrec *left, *right;
} REL;

typedef struct queryrec
{
    char	*word;
    REL		*rel;
    int		numrel;
    struct queryrec *left, *right;
} QUERY;

QUERY	*qrys;
char	*buf;
int	buflen;

void	loadrel();
char	*getword(char **);
QUERY	*findqry(QUERY *,char *), *insqry(QUERY *,char *);
REL	*findrel(REL *,char *), *insrel(REL *,char *);
char	*fgets(char *,int,FILE *);
char	*malloc(int);

main(int argc, char *argv[])
{
    int		actq;	    /* how many queries investigated */
    int		dec, j, k, nr, pos, target;
    double	avg, avgs[11], pprec[11];
    char	*ptr, *p;
    int		numans;			/* answers so far */
    double	avprec[3];		/* average precisions at 10, 20, 30 */
    double	avrel;
    QUERY	*q;
    REL		*r;

    if( argc != 2 )
    {
	fprintf(stderr, "Usage: wordrp relfile\n");
	exit(1);
    }

    loadrel(argv[1]);
    buflen = BUFLEN * sizeof(char);
    buf = (char *) malloc(buflen);

    for( j=0 ; j<=10 ; j++ )
	avgs[j] = 0.0;

    avrel = avprec[0] = avprec[1] = avprec[2] = 0.0;

    for( actq=0 ; fgets(buf, BUFLEN, stdin) != NULL ; )
    {
	p = buf;
	ptr = getword(&p);
	q = findqry(qrys, ptr);

	if( q == NULL || q->numrel == 0 ) /* no information for this query */
	    continue;

	actq++;

	numans = 0;

	    /* Compute 11-pt avg as per Chris Buckley */

	for( dec=0 ; dec<=10 ; dec++ )
	    pprec[dec] = 0.0;

	ptr = getword(&p);
	    /* I believe this WAS a bug.
	       With the test for ptr != NULL, the loop would terminate
	       early and not carry on, so that even if all relevant
	       terms had been found the later precisions would have
	       recall 0. */
	for( dec=nr=pos=0 ; dec < 10 /* && ptr != NULL */ ; dec++ )
	{
	    target = ( (dec+1)*q->numrel + 9 )/10;
		/* scan buf for relevant words */

/* printf("target %d, nr %d\n", target, nr); */

	    if( ptr != NULL )
	    {
		for( ; nr <= target && nr < q->numrel && ptr != NULL ; pos++ )
		{
		    if( findrel(q->rel, ptr) )
		    {
			nr++;
			for( j=dec ; j>=0 ; j-- )
			{
			    if( (double)nr/(pos+1.0) > pprec[j] )
				pprec[j] = (double)nr/(pos+1.0);
			    else
				break;
			}
		    }
		    numans++;
		    if( numans==10 )
			avprec[0] += nr/10.0;
		    else if( numans==20 )
			avprec[1] += nr/20.0;
		    else if( numans==30 )
			avprec[2] += nr/30.0;
		    ptr = getword(&p);
		}
		if( pos > 0 )
		{
		    pprec[dec+1] = (double)nr/(double)pos;
		}
	    }
	    else if( nr >= target )
	    {
		for( j=dec ; j>=0 ; j-- )
		{
		    if( (double)nr/(pos+1.0) > pprec[j] )
			pprec[j] = (double)nr/(pos+1.0);
		    else
			break;
		}
		pprec[dec+1] = (double)nr/(double)pos;
	    }

/* for(j=0 ; j<=dec+1 ; j++ ) printf("%1.3f  ",  pprec[dec+1]); printf("\n"); */

	}

	avrel += (double)nr;
	if( numans<10 )
	    avprec[0] += nr/10.0;
	if( numans<20 )
	    avprec[1] += nr/20.0;
	if( numans<30 )
	    avprec[2] += nr/30.0;

	for( dec=0 ; dec <= 10 ; dec++ )
	    avgs[dec] += pprec[dec];
    }

    for( avg=0.0, dec=0 ; dec<=10 ; dec++ )
	avg += avgs[dec];

    printf("r/p=%1.1f p@10=%1.1f p@20=%1.1f p@30=%1.1f rel=%1.1f\n",
	100.0*avg/(actq*11),
	100.0*avprec[0]/actq, 100.0*avprec[1]/actq, 100.0*avprec[2]/actq,
	avrel/actq);

    exit(0);
}


/* Read relevance judgements from file, insert into binary tree */
void
loadrel(char *file)
{
    FILE	*fp, *fopen();
    char	buf[BUFLEN], *ptr, *p;
    QUERY	*q;
    REL		*r;

    if( ( fp = fopen(file, "r") ) == NULL )
    {
	fprintf(stderr, "Error: can't open %s for reading.\n", file);
	exit(1);
    }

    qrys = NULL;
    while( fgets(buf, BUFLEN, fp) != NULL )
    {
	p = buf;
	ptr = getword(&p);		/* query term */
	qrys = insqry(qrys, ptr);	/* yukko, double traverse */
	q = findqry(qrys, ptr);
	for( ptr=getword(&p) ; ptr != NULL ; ptr=getword(&p) )
	{
	    r = insrel(q->rel, ptr);
	    q->numrel++;
	    if( q->rel == NULL )
		q->rel = r;
	}
    }
}


/* Insert str into binary tree q */
QUERY *
insqry(QUERY *q, char *str)
{
    int		val;

    if( q == NULL )
    {
	q = (QUERY *) malloc(sizeof(QUERY));
	q->word = malloc(strlen(str) + 1);
	strcpy(q->word, str);
	q->left = q->right = NULL;
	q->rel = NULL;
	q->numrel = 0;
    }
    else if( ( val = strcmp(str, q->word) ) == 0 )
	fprintf(stderr, "Warning, repeated query %s\n", str);
    else if( val < 0 )
	q->left = insqry(q->left, str);
    else
	q->right = insqry(q->right, str);

    return(q);
}


/* Find str in binary tree q */
QUERY *
findqry(QUERY *q, char *str)
{
    int		val;

    if( q == NULL )
	return(NULL);
    else if( ( val = strcmp(str, q->word) ) == 0 )
	return(q);
    else if( val < 0 )
	return(findqry(q->left, str));
    else
	return(findqry(q->right, str));
}


/* Insert str into binary tree r */
REL *
insrel(REL *r, char *str)
{
    int		val;

    if( r == NULL )
    {
	r = (REL *) malloc(sizeof(REL));
	r->word = malloc(strlen(str) + 1);
	strcpy(r->word, str);
	r->left = r->right = NULL;
    }
    else if( ( val = strcmp(str, r->word) ) == 0 )
	fprintf(stderr, "Warning, repeated relevance judgement %s\n", str);
    else if( val < 0 )
	r->left = insrel(r->left, str);
    else
	r->right = insrel(r->right, str);
    return(r);
}


/* Find str in binary tree r */
REL *
findrel(REL *r, char *str)
{
    int		val;

    if( r == NULL )
	return(NULL);
    else if( ( val = strcmp(str, r->word) ) == 0 )
	return(r);
    else if( val < 0 )
	return(findrel(r->left, str));
    else
	return(findrel(r->right, str));
}


/* Parse a word out of current string; assumes newline+null terminator.
   The argument is updated to point to the start of the next string, or
   to '\0' at the end. */
char *
getword(char **str)
{
    char	*ptr;

    if( **str == '\0' )
	return(NULL);

    ptr = *str;
    while( **str != ' ' && **str != '\n' )
    {
	if( **str == '\0' )
	{
	    fprintf(stderr, "Badly terminated string %s\n", ptr);
	    exit(1);
	}
	(*str)++;
	if( (int)*str >= (int)buf+buflen )
	{
	    fprintf(stderr, "Line too long\n");
	    exit(1);
	}
    }
    **str = '\0';
    (*str)++;
    while( **str == ' ' )
    {
	if( (int)*str >= (int)buf+buflen )
	{
	    fprintf(stderr, "Line too long\n");
	    exit(1);
	}
	(*str)++;
    }

    return(ptr);
}
